unit midi_defs;

interface

uses

  System.SysUtils, OSCif;

const

  MIDI_CH_OMNI = 16;           // MIID channels are zero relative, 0 -> ch1, 15 -> ch 16 and 16 -> any channel / omni

  // Midi command definitions

  CNoteOff              = $80; // Length = 3;
  CNoteOn               = $90; // Length = 3;
  CKeyPressure          = $a0; // Length = 3;
  CControlChange        = $b0; // Length = 3;
  CProgramChange        = $c0; // Length = 2;
  CChannelPressure      = $d0; // Length = 2;
  CPitchBend            = $e0; // Length = 2;
  CSysEx                = $f0; // Length = n;

  // When aCmd And $f0 = CSysEx low nibble of 1st byte is :

  SSysEx                = $00; // Length = 1, followed by arbitrary amt of data, finalized by SEox;
  SSongPosition         = $02; // Length = 3;
  SSongSelect           = $03; // Length = 2;
  STuneRequest          = $06; // Length = 1;
  SEox                  = $07; // Length = 1;
  STimingClock          = $08; // Length = 1;
  SStart                = $0a; // Length = 1;
  SContinue             = $0b; // Length = 1;
  SStop                 = $0c; // Length = 1;
  SActiveSense          = $0e; // Length = 1;
  SReset                = $0f; // Length = 1;

  // Pseudo CC's

  CCAllSoundOff         = 120; // Pseudo CC - value should be 0
  CCResetAllControllers = 121; // Pseudo CC - value should be 0
  CCLoccalControl       = 122; // Pseudo CC - value 0 = off, value 127 = on
  CCAllNotesOff         = 123; // Pseudo CC - value should be 0
  CCOmniModeOff         = 124; // Pseudo CC - value should be 0
  CCOmniModeOn          = 125; // Pseudo CC - value should be 0
  CCMonoModeOn          = 126; // Pseudo CC - value 0 = omni on, other values M = omni off with M channels used
  CCPolyModeOn          = 127; // Pseudo CC - value should be 0

  // Meta types

  CMSequenceNumber      = $00;
  CMTextEvent           = $01;
  CMCopyRightNotice     = $02;
  CMSequenceTrackName   = $03;
  CMInstrumentName      = $04;
  CMLyric               = $05;
  CMMarker              = $06;
  CMCuePoint            = $07;
  CMMidiChannelPrefix   = $20;
  CMEndOfTrack          = $2f;
  CMSetTempo            = $51;
  CMSMTPOffset          = $54;
  CMTimeSignature       = $58;
  CMKeySignature        = $59;
  CMSequencerSpecific   = $7F;

type

  TInts = array of Integer;

  TMidiMessage = record
    Command : Byte;
    Channel : Byte;
    Data    : TInts;
  end;

  TShortMidiMessage = record
    DataLen : Byte;
    Command : Byte;
    Channel : Byte;
    Data1   : Byte;
    Data2   : Byte;
  end;

  TMidiRxState = (
    mrxNormal,
    mrxRunningStatus
  );

  TOnMidiMessage      = procedure( const aSender: TObject; const aMsg: TMidiMessage     ) of object;
  TOnShortMidiMessage = procedure( const aSender: TObject; const aMsg: TShortMidiMessage) of object;

  TMidiReceiver = class
  private
    FOnMessage     : TOnMidiMessage;
    FRxState       : TMidiRxState;
    FDataCount     : Integer;
    FRunningStatus : Byte;
    FChannel       : Byte;
    FData          : Byte;
    FSysEx         : TBytes;
  private
    procedure   AcceptCommand( aByte: Byte);
    procedure   AcceptData   ( aByte: Byte);
    procedure   AcceptSysex  ( aByte: Byte);
    procedure   Reset;
    procedure   ProcessNoteOff        ( aVelocity, aNote, aChannel: Byte);
    procedure   ProcessNoteOn         ( aVelocity, aNote, aChannel: Byte);
    procedure   ProcessKeyPressure    ( aPressure, aNote, aChannel: Byte);
    procedure   ProcessControlChange  ( aValue, aController, aChannel: Byte);
    procedure   ProcessProgramChange  ( aProgram, aChannel: Byte);
    procedure   ProcessChannelPressure( aValue, aChannel: Byte);
    procedure   ProcessPitchBend      ( aData: Word; aChannel: Byte);
    procedure   ProcessStartSysEx;
    procedure   ProcessEndSysEx;
    procedure   ProcessSongPosition   ( aData: Word);
    procedure   ProcessSongSelect     ( aSong: Byte);
    procedure   ProcessTuneRequest;
    procedure   ProcessTimingClock;
    procedure   ProcessStart;
    procedure   ProcessContinue;
    procedure   ProcessStop;
    procedure   ProcessActiveSense;
    procedure   ProcessReset;
    procedure   ProcessSysExData      ( aData: Byte);
    procedure   ProcessMessage( const aMsg: TMidiMessage);
  public
    procedure   AcceptByte( aByte: Byte);
    procedure   AcceptBytes( aBytes: TBytes);
  public
    property    OnMessage: TOnMidiMessage read FOnMessage write FOnMessage;
  end;


  procedure MidiMessageToBytes     ( const aMsg: TMidiMessage     ; var aData: TBytes);
  procedure ShortMidiMessageToBytes( const aMsg: TShortMidiMessage; var aData: TBytes);
  function  MidiMessageToLog       ( const aMsg: TMidiMessage     ): string;
  function  ShortMidiMessageToLog  ( const aMsg: TShortMidiMessage): string;


implementation


  function  MakeMidiNoteOff( aVelocity, aNote, aChannel: Byte): TMidiMessage;
  begin
    Result.Command := CNoteOff;
    Result.Channel := aChannel;
    SetLength( Result.Data, 2);
    Result.Data[ 0] := aNote;
    Result.Data[ 1] := aVelocity;
  end;

  function  MakeMidiNoteOn( aVelocity, aNote, aChannel: Byte): TMidiMessage;
  begin
    Result.Command := CNoteOn;
    Result.Channel := aChannel;
    SetLength( Result.Data, 2);
    Result.Data[ 0] := aNote;
    Result.Data[ 1] := aVelocity;
  end;

  function  MakeMidiKeyPressure( aPressure, aNote, aChannel: Byte): TMidiMessage;
  begin
    Result.Command := CKeyPressure;
    Result.Channel := aChannel;
    SetLength( Result.Data, 2);
    Result.Data[ 0] := aNote;
    Result.Data[ 1] := aPressure;
  end;

  function  MakeMidiControlChange( aValue, aController, aChannel: Byte): TMidiMessage;
  begin
    Result.Command := CControlChange;
    Result.Channel := aChannel;
    SetLength( Result.Data, 2);
    Result.Data[ 0] := aController;
    Result.Data[ 1] := aValue;
  end;

  function  MakeMidiProgramChange( aProgram, aChannel: Byte): TMidiMessage;
  begin
    Result.Command := CProgramChange;
    Result.Channel := aChannel;
    SetLength( Result.Data, 1);
    Result.Data[ 0] := aProgram;
  end;

  function  MakeMidiChannelPressure( aValue, aChannel: Byte): TMidiMessage;
  begin
    Result.Command := CChannelPressure;
    Result.Channel := aChannel;
    SetLength( Result.Data, 1);
    Result.Data[ 0] := aValue;
  end;

  function  MakeMidiPitchBend( aData: Word; aChannel: Byte): TMidiMessage;
  begin
    Result.Command := CPitchBend;
    Result.Channel := aChannel;
    SetLength( Result.Data, 1);
    Result.Data[ 0] := aData;
  end;

  function  MakeMidiSongPosition( aData: Word): TMidiMessage;
  begin
    Result.Command := SSongPosition;
    Result.Channel := 0;
    SetLength( Result.Data, 1);
    Result.Data[ 0] := aData;
  end;

  function  MakeMidiSongSelect( aSong: Byte): TMidiMessage;
  begin
    Result.Command := SSongSelect;
    Result.Channel := 0;
    SetLength( Result.Data, 1);
    Result.Data[ 0] := aSong;
  end;

  function  MakeMidiTuneRequest: TMidiMessage;
  begin
    Result.Command := STuneRequest;
    Result.Channel := 0;
    SetLength( Result.Data, 0);
  end;

  function  MakeMidiTimingClock: TMidiMessage;
  begin
    Result.Command := STimingClock;
    Result.Channel := 0;
    SetLength( Result.Data, 0);
  end;

  function  MakeMidiStart: TMidiMessage;
  begin
    Result.Command := SStart;
    Result.Channel := 0;
    SetLength( Result.Data, 0);
  end;

  function  MakeMidiContinue: TMidiMessage;
  begin
    Result.Command := SContinue;
    Result.Channel := 0;
    SetLength( Result.Data, 0);
  end;

  function  MakeMidiStop: TMidiMessage;
  begin
    Result.Command := SStop;
    Result.Channel := 0;
    SetLength( Result.Data, 0);
  end;

  function  MakeMidiActiveSense: TMidiMessage;
  begin
    Result.Command := SActiveSense;
    Result.Channel := 0;
    SetLength( Result.Data, 0);
  end;

  function  MakeMidiReset: TMidiMessage;
  begin
    Result.Command := SReset;
    Result.Channel := 0;
    SetLength( Result.Data, 0);
  end;

  function  MakeMidiSysExData( const aData: TBytes): TMidiMessage;
  var
    i : Integer;
  begin
    Result.Command := SSysEx;
    Result.Channel := 0;
    SetLength( Result.Data, Length( aData));
    for i := 0 to Length( Result.Data) - 1
    do Result.Data[ i] := aData[ i];
  end;

  procedure MidiMessageToBytes( const aMsg: TMidiMessage; var aData: TBytes);
  begin
    SetLength( aData, 0);

    case aMsg.Command of

      CNoteOff,          // = $80; // Length = 3;
      CNoteOn,           // = $90; // Length = 3;
      CKeyPressure,      // = $a0; // Length = 3;
      CControlChange :   // = $b0; // Length = 3;
        begin
          if Length( aMsg.Data) = 2
          then begin
            SetLength( aData, 3);
            aData[ 0] := ( aMsg.Command and $f0) or ( aMsg.Channel and $0f);
            aData[ 1] := aMsg.Data[ 0];
            aData[ 2] := aMsg.Data[ 1];
          end;
        end;

      CProgramChange,    // = $c0; // Length = 2;
      CChannelPressure : // = $d0; // Length = 2;
        begin
          if Length( aMsg.Data) = 2
          then begin
            SetLength( aData, 2);
            aData[ 0] := ( aMsg.Command and $f0) or ( aMsg.Channel and $0f);
            aData[ 1] := aMsg.Data[ 0];
          end;
        end;

      CPitchBend :       // = $e0; // Length = 3;
        begin
          if Length( aMsg.Data) = 2
          then begin
            SetLength( aData, 3);
            aData[ 0] := ( aMsg.Command  and $00f0) or ( aMsg.Channel and $000f);
            aData[ 1] :=   aMsg.Data[ 0] and $007f;
            aData[ 2] := ( aMsg.Data[ 0] and $3f80) shr 7;
          end;
        end;

//      CSysEx :          // = $f0; // Length = n;
//        begin
//        end;

    end;
  end;

  procedure ShortMidiMessageToBytes( const aMsg: TShortMidiMessage; var aData: TBytes);
  begin
    SetLength( aData, 0);

    case aMsg.Command of

      CNoteOff,          // = $80; // Length = 3;
      CNoteOn,           // = $90; // Length = 3;
      CKeyPressure,      // = $a0; // Length = 3;
      CControlChange :   // = $b0; // Length = 3;
        begin
          if aMsg.DataLen = 2
          then begin
            SetLength( aData, 3);
            aData[ 0] := ( aMsg.Command and $f0) or ( aMsg.Channel and $0f);
            aData[ 1] := aMsg.Data1;
            aData[ 2] := aMsg.Data2;
          end;
        end;

      CProgramChange,    // = $c0; // Length = 2;
      CChannelPressure : // = $d0; // Length = 2;
        begin
          if aMsg.DataLen = 1
          then begin
            SetLength( aData, 2);
            aData[ 0] := ( aMsg.Command and $f0) or ( aMsg.Channel and $0f);
            aData[ 1] := aMsg.Data1;
          end;
        end;

      CPitchBend :       // = $e0; // Length = 3;
        begin
          if aMsg.DataLen = 2
          then begin
            SetLength( aData, 3);
            aData[ 0] := ( aMsg.Command  and $00f0) or ( aMsg.Channel and $000f);
            aData[ 1] :=   aMsg.Data1 and $007f;
            aData[ 2] := ( aMsg.Data2 and $3f80) shr 7;
          end;
        end;

//      CSysEx :          // = $f0; // Length = n;
//        begin
//        end;

    end;
  end;


  function MidiCmdToStr( aCmd: Byte): string;
  begin
    case aCmd of
      CNoteOff         : Result := 'note off';         // = $80; // Length = 3;
      CNoteOn          : Result := 'note on';          // = $90; // Length = 3;
      CKeyPressure     : Result := 'key pressure';     // = $a0; // Length = 3;
      CControlChange   : Result := 'control change';   // = $b0; // Length = 3;
      CProgramChange   : Result := 'program change';   // = $c0; // Length = 2;
      CChannelPressure : Result := 'channel pressure'; // = $d0; // Length = 2;
      CPitchBend       : Result := 'pitch bend';       // = $e0; // Length = 2;
      CSysEx           : Result := 'sysex';            // = $f0; // Length = n;
    end;
  end;


  function  MidiMessageToLog( const aMsg: TMidiMessage): string;
  var
    i : Integer;
  begin
    Result := Format( 'ch %d cmd %.2x [%s] ', [ aMsg.Channel + 1, aMsg.Command, MidiCmdToStr( aMsg.Command)]);

    for i := 0 to Length( aMsg.Data) - 1
    do Result := Result + Format( '%.2x ', [ aMSg.Data[ i]]);

    SetLength( Result, Length( Result) - 1);
  end;


  function  ShortMidiMessageToLog( const aMsg: TShortMidiMessage): string;
  begin
    Result := Format( 'ch %d cmd %.2x [%s]', [ aMsg.Channel + 1, aMsg.Command, MidiCmdToStr( aMsg.Command)]);

    if aMsg.DataLen >= 1
    then Result := Result + Format( ' %.2x', [ aMSg.Data1]);

    if aMsg.DataLen = 2
    then Result := Result + Format( ' %.2x', [ aMSg.Data2]);
  end;


{ ========
  TMidiReceiver = class
  private
    FOnMidiMessage : TOnMidiMessage;
  public
    property    OnMidiMessage: TOnMidiMessage read FOnMidiMessage write FOnMidiMessage;
  private
}

    procedure   TMidiReceiver.AcceptCommand( aByte: Byte);
    begin
      FDataCount := 0;

      case aByte and $f0 of

        CSysEx : AcceptSysex( aByte);

        else begin
          FRunningStatus := aByte;
          FRxState       := mrxRunningStatus;
          FChannel       := aByte and $0f;
        end;

      end;
    end;

    procedure   TMidiReceiver.AcceptData( aByte: Byte);
    begin
      case FRunningStatus and $f0 of

        CNoteOff :         // = $80; // Length = 3;
          begin
            if FDataCount = 1
            then begin
              ProcessNoteOff( aByte, FData, FChannel);
              FDataCount := 0;
            end
            else begin
              FData := aByte;
              Inc( FDataCount);
            end;
          end;

        CNoteOn :          // = $90; // Length = 3;
          begin
            if FDataCount = 1
            then begin
              ProcessNoteOn( aByte, FData, FChannel);
              FDataCount := 0;
            end
            else begin
              FData := aByte;
              Inc( FDataCount);
            end;
          end;

        CKeyPressure :     // = $a0; // Length = 3;
          begin
            if FDataCount = 1
            then begin
              ProcessKeyPressure( aByte, FData, FChannel);
              FDataCount := 0;
            end
            else begin
              FData := aByte;
              Inc( FDataCount);
            end;
          end;

        CControlChange :   // = $b0; // Length = 3;
          begin
            if FDataCount = 1
            then begin
              ProcessControlChange( aByte, FData, FChannel);
              FDataCount := 0;
            end
            else begin
              FData := aByte;
              Inc( FDataCount);
            end;
          end;

        CProgramChange :   // = $c0; // Length = 2;
          begin
            ProcessProgramChange( aByte,FChannel);
          end;

        CChannelPressure : // = $d0; // Length = 2;
          begin
            ProcessChannelPressure( aByte,FChannel);
          end;

        CPitchBend :       // = $e0; // Length = 2;
          begin
            if FDataCount = 1
            then begin
              ProcessPitchBend(( aByte and $7f) + 128 * ( FData and $7f), FChannel);
              FDataCount := 0;
            end
            else begin
              FData := aByte;
              Inc( FDataCount);
            end;
          end;

        CSysEx :           // = $f0; // Length = n;
          begin
            ProcessSysExData( aByte);
          end;

        SSongPosition :
          begin
            if FDataCount = 1
            then begin
              ProcessSongPosition(( aByte and $7f) + 128 * ( FData and $7f));
              FDataCount := 0;
            end
            else begin
              FData := aByte;
              Inc( FDataCount);
            end;
          end;

        SSongSelect :
          begin
            ProcessSongSelect( aByte);
          end;

        else Reset;

      end;
    end;

    procedure   TMidiReceiver.AcceptSysex( aByte: Byte);
    begin
      case aByte and $0f of

        SSysEx :          // = $00; // Length = 1, followed by arbitrary amt of data, finalized by SEox;
          begin
            FRunningStatus := SSysEx;
            FRxState       := mrxRunningStatus;
            ProcessStartSysEx;
          end;

        SSongPosition :   // = $02; // Length = 3;
          begin
            FRunningStatus := SSongPosition;
            FRxState       := mrxRunningStatus;
          end;

        SSongSelect :     // = $03; // Length = 2;
          begin
            FRunningStatus := SSongPosition;
            FRxState       := mrxRunningStatus;
          end;

        STuneRequest :    // = $06; // Length = 1;
          begin
            ProcessTuneRequest;
          end;

        SEox :            // = $07; // Length = 1;
          begin
            ProcessEndSysEx;
          end;

        STimingClock :    // = $08; // Length = 1;
          begin
            ProcessTimingClock;
          end;

        SStart :          // = $0a; // Length = 1;
          begin
            ProcessStart;
          end;

        SContinue :       // = $0b; // Length = 1;
          begin
            ProcessContinue;
          end;

        SStop :           // = $0c; // Length = 1;
          begin
            ProcessStop;
          end;

        SActiveSense :    // = $0e; // Length = 1;
          begin
            ProcessActiveSense;
          end;

        SReset :          // = $0f; // Length = 1;
          begin
            ProcessReset;
          end;

      end;
    end;

    procedure   TMidiReceiver.Reset;
    begin
      FRxState       := mrxNormal;
      FRunningStatus := 0;
    end;

    procedure   TMidiReceiver.ProcessNoteOff( aVelocity, aNote, aChannel: Byte);
    begin
      ProcessMessage( MakeMidiNoteOff( aVelocity, aNote, aChannel));
    end;

    procedure   TMidiReceiver.ProcessNoteOn( aVelocity, aNote, aChannel: Byte);
    begin
      if aVelocity = 0
      then ProcessMessage( MakeMidiNoteOff( 0        , aNote, aChannel))
      else ProcessMessage( MakeMidiNoteOn ( aVelocity, aNote, aChannel));
    end;

    procedure   TMidiReceiver.ProcessKeyPressure( aPressure, aNote, aChannel: Byte);
    begin
      ProcessMessage( MakeMidiKeyPressure( aPressure, aNote, aChannel));
    end;

    procedure   TMidiReceiver.ProcessControlChange( aValue, aController, aChannel: Byte);
    begin
      ProcessMessage( MakeMidiControlChange( aValue, aController, aChannel));
    end;

    procedure   TMidiReceiver.ProcessProgramChange( aProgram, aChannel: Byte);
    begin
      ProcessMessage( MakeMidiProgramChange( aProgram, aChannel));
    end;

    procedure   TMidiReceiver.ProcessChannelPressure( aValue, aChannel: Byte);
    begin
      ProcessMessage( MakeMidiChannelPressure( aValue, aChannel));
    end;

    procedure   TMidiReceiver.ProcessPitchBend( aData: Word; aChannel: Byte);
    begin
      ProcessMessage( MakeMidiPitchBend( aData, aChannel));
    end;

    procedure   TMidiReceiver.ProcessStartSysEx;
    begin
      SetLength( FSysex, 0);
    end;

    procedure   TMidiReceiver.ProcessEndSysEx;
    begin
      // Sysex message completed and stored in FMidiSysex
      ProcessMessage( MakeMidiSysExData( FSysex));
    end;

    procedure   TMidiReceiver.ProcessSongPosition( aData: Word);
    begin
      ProcessMessage( MakeMidiSongPosition( aData));
    end;

    procedure   TMidiReceiver.ProcessSongSelect( aSong: Byte);
    begin
      ProcessMessage( MakeMidiSongSelect( aSong));
    end;

    procedure   TMidiReceiver.ProcessTuneRequest;
    begin
      ProcessMessage( MakeMidiTuneRequest);
    end;

    procedure   TMidiReceiver.ProcessTimingClock;
    begin
      ProcessMessage( MakeMidiTimingClock);
    end;

    procedure   TMidiReceiver.ProcessStart;
    begin
      ProcessMessage( MakeMidiStart);
    end;

    procedure   TMidiReceiver.ProcessContinue;
    begin
      ProcessMessage( MakeMidiContinue);
    end;

    procedure   TMidiReceiver.ProcessStop;
    begin
      ProcessMessage( MakeMidiStop);
    end;

    procedure   TMidiReceiver.ProcessActiveSense;
    begin
      ProcessMessage( MakeMidiActiveSense);
    end;

    procedure   TMidiReceiver.ProcessReset;
    begin
      ProcessMessage( MakeMidiReset);
    end;

    procedure   TMidiReceiver.ProcessSysExData( aData: Byte);
    begin
      if Length( FSysex) >= 1024 // Prevent buffer from getting arbitrary large
      then begin
        ProcessEndSysEx;         // process the partial message
        ProcessStartSysEx;       // and start a new one
      end;
      SetLength( FSysex, Length( FSysex) + 1);
      FSysex[ Length( FSysex) - 1] := aData;
    end;

    procedure   TMidiReceiver.ProcessMessage( const aMsg: TMidiMessage);
    begin
      if Assigned( FOnMessage)
      then FOnMessage( Self, aMsg);
    end;

//  public

    procedure   TMidiReceiver.AcceptByte( aByte: Byte);
    begin
      case FRxState of

        mrxNormal :
          begin
            if aByte and $80 = $80
            then AcceptCommand( aByte);
          end;

        mrxRunningStatus :
          begin
            if aByte and $80 = $80
            then AcceptCommand( aByte)
            else AcceptData   ( aByte);
          end;

      end;
    end;


    procedure   TMidiReceiver.AcceptBytes( aBytes: TBytes);
    var
      i : Integer;
    begin
      for i := 0 to Length( aBytes) - 1
      do AcceptByte( aBytes[ i]);
    end;



end.

